/* $Id: authserv.C,v 1.33 1999/05/28 21:50:27 kaminsky Exp $ */

/*
 *
 * Copyright (C) 1999 Michael Kaminsky (kaminsky@lcs.mit.edu)
 * Copyright (C) 1998 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "authserv.h"
#include "parseopt.h"
#include "rxx.h"

void
unixaccept (ptr<axprt_stream> x, const authunix_parms *aup)
{
  if (x)
    vNew authclient (x, aup);
}

ptr<axprt_stream>
cloneaccept (int fd)
{
  if (fd < 0)
    fatal ("EOF from authserv\n");
  tcp_nodelay (fd);
  ref<axprt_crypt> x = axprt_crypt::alloc (fd);
  vNew authclient (x);
  return x;
}

static void
parseconfig (str cf)
{
  parseargs pa (cf);
  bool errors = false;

  int line;
  vector<str> av;

  while (pa.getline (&av, &line)) {
    if (!strcasecmp (av[0], "srpfile")) {
      if (srpglobal) {
	errors = true;
	warn << cf << ":" << line << ": srpfile already specified\n";
      }
      else if (av.size () == 2) {
	str key = file2wstr (av[1]);
	if (!key) {
	  errors = true;
	  warn << av[1] << ": " << strerror (errno) << "\n";
	  warn << cf << ":" << line << ": could not read srpfile\n";
	}
	else if (!(srpglobal = import_srp_params (key))) {
	  errors = true;
	  warn << cf << ":" << line << ": could not decode srpfile\n";
	}
      }
      else {
	errors = true;
	warn << cf << ":" << line << ": usage: srpfile path\n";
      }
    }
    else if (!strcasecmp (av[0], "keyfile")) {
      if (authsrvkey) {
	  errors = true;
	  warn << cf << ":" << line << ": keyfile already specified\n";
      }
      else if (av.size () == 2) {
	str key = file2wstr (av[1]);
	if (!key) {
	  errors = true;
	  warn << av[1] << ": " << strerror (errno) << "\n";
	  warn << cf << ":" << line << ": could not read keyfile\n";
	}
	else if (!(authsrvkey = import_rabin_priv (key, NULL))) {
	  errors = true;
	  warn << cf << ":" << line << ": could not decode keyfile\n";
	}
      }
      else {
	errors = true;
	warn << cf << ":" << line << ": usage: keyfile path\n";
      }
    }
    else if (!strcasecmp (av[0], "shareduserfile")) {
      if (shareduserfile) {
	  errors = true;
	  warn << cf << ":" << line << ": shareduserfile already specified\n";
      }
      else if (av.size () == 2) {
	shareduserfile = New refcounted<userfile> (av[1], true);
      }
      else {
	errors = true;
	warn << cf << ":" << line << ": usage: shareduserfile path\n";
      }
    }
    else if (!strcasecmp (av[0], "denyfile")) {
      if (denyfile) {
	  errors = true;
	  warn << cf << ":" << line << ": denyfile already specified\n";
      }
      else if (av.size () == 2) {
	if (!file2str (av[1])) {
	  errors = true;
	  warn << av[1] << ": " << strerror (errno) << "\n";
	  warn << cf << ":" << line << ": could not read denyfile\n";
	}
	else
	  denyfile = av[1];
      }
      else {
	errors = true;
	warn << cf << ":" << line << ": usage: denyfile path\n";
      }
    }
    else {
      errors = true;
      warn << cf << ":" << line << ": Unknown directive '"
	   << av[0] << "'\n";
    }
  }

  if (!errors && !authsrvkey) {
    str keyfile = sfsconst_etcfile ("sfs_host_key");
    if (!keyfile) {
      errors = true;
      warn << "cannot locate default file sfs_host_key\n";
    }
    else {
      str key = file2wstr (keyfile);
      if (!key) {
	errors = true;
	warn << keyfile << ": " << strerror (errno) << "\n";
      }
      else if (!(authsrvkey = import_rabin_priv (key, NULL))) {
	errors = true;
	warn << "could not decode " << keyfile << "\n";
      }
    }
  }
  if (!errors && !srpglobal) {
    str keyfile = sfsconst_etcfile ("sfs_srp_parms");
    if (!keyfile) {
      errors = true;
      warn << "cannot locate default file sfs_srp_parms\n";
    }
    else {
      str key = file2wstr (keyfile);
      if (!key) {
	errors = true;
	warn << keyfile << ": " << strerror (errno) << "\n";
      }
      else if (!(srpglobal = import_srp_params (key))) {
	errors = true;
	warn << "could not decode " << keyfile << "\n";
      }
    }
  }
  if (errors)
    fatal << "errors in configuration file\n";

  if (!denyfile)
    denyfile = sfsconst_etcfile ("sfs_deny");

  hinfo.type = SFS_HOSTINFO;
  if (!(hinfo.hostname = myname ()))
    fatal << "could not figure out my host name\n";
  hinfo.pubkey = authsrvkey->n;
  if (!sfs_mkhostid (&hostid, hinfo))
    fatal << "could not marshall my own hostinfo\n";
}

static void
usage ()
{
  warnx << "usage: " << progname 
	<< " [-f configfile] [{-r | -w} userfile]\n";
  exit (1);
}

extern char *optarg;
extern int optind;

int
main (int argc, char **argv)
{
  enum { minsize = 768 };

  str configfile;

  setprogname (argv[0]);

  int ch;
  while ((ch = getopt (argc, argv, "f:r:w:")) != -1)
    switch (ch) {
    case 'r':
    case 'w':
      userfiles.push_back (userfile (optarg, ch == 'w'));
      break;
    case 'f':
      configfile = optarg;
      break;
    case '?':
    default:
      usage ();
    }

  argc -= optind;
  argv += optind;
  if (argc > 1)
    usage ();

  sfsconst_init ();

  setgid (sfs_gid);
  setgroups (0, NULL);

  warn ("pid %d\n", getpid ());

  if (!configfile)
    configfile = sfsconst_etcfile_required ("authserv_config");
  parseconfig (configfile);

  if (!builddir)
    random_set_seedfile (strbuf ("%s/random_seed", sfsdir.cstr ()));
  random_init ();

  if (userfiles.empty ()) {
    str uf = sfsconst_etcfile ("sfs_users");
    if (!uf)
      uf = strbuf ("%s/%s", etc1dir, "sfs_users");
    userfiles.push_back (userfile (uf, true));
  }

  if (!authdbinit ())
    fatal << "Could not initialize authentication database\n";

  warn << "serving /sfs/" << hinfo.hostname << ":"
       << armor32 (str (hostid.base (), sizeof (sfs_hash))) << "\n";

  sfs_suidserv ("authserv", wrap (unixaccept));
  if (!cloneserv (0, wrap (cloneaccept)))
    warn ("No sfssd detected, only accepting unix-domain clients\n");

  amain ();
}
