/* $Id: sfsctl.C,v 1.10 1999/04/18 23:02:51 kaminsky Exp $ */

/*
 *
 * Copyright (C) 1999 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "sfsrwcd.h"

#define SFSPREF ".SFS "
#define SFSFH SFSPREF "FH"
#define SFSFS SFSPREF "FS"

const u_int32_t srvno_fh = 0;
const u_int32_t srvno_fs = 1;

struct fileinfo {
  const str fspath;
  const nfs_fh3 fh;

  fileinfo (server *s, const nfs_fh3 &fh);
};

struct sfsctl {
  ptr<asrv> s;
  const sfs_aid aid;
  int32_t pid;
  struct authunix_parms aup;
  ihash_entry<sfsctl> hlink;

  ptr<fileinfo> fip;

  sfsctl (ref<axprt_stream> x, const authunix_parms *aup);
  ~sfsctl ();
  void setpid (int32_t npid);

  void dispatch (svccb *sbp);
};

ihash2<const sfs_aid, int32_t, sfsctl,
  &sfsctl::aid, &sfsctl::pid, &sfsctl::hlink> ctltab;

fileinfo::fileinfo (server *s, const nfs_fh3 &ffh)
  : fspath (s->path), fh (ffh)
{
}

void
sfsctl::setpid (int32_t npid)
{
  ctltab.remove (this);
  pid = npid;
  ctltab.insert (this);
}

sfsctl::sfsctl (ref<axprt_stream> x, const authunix_parms *naup)
  : aid (aup2aid (naup)), pid (0)
{
  const int glen = naup->aup_len * sizeof (naup->aup_gids[0]);
  aup = *naup;
  aup.aup_machname = xstrdup (naup->aup_machname);
#ifdef AUTHUNIX_GID_T
  aup.aup_gids = (gid_t *) xmalloc (glen);
#else /* !AUTHUNIX_GID_T */
  aup.aup_gids = (int *) xmalloc (glen);
#endif /* !AUTHUNIX_GID_T */
  memcpy (aup.aup_gids, naup->aup_gids, glen);
  s = asrv::alloc (x, sfsctl_prog_1, wrap (this, &sfsctl::dispatch));
  ctltab.insert (this);
}

sfsctl::~sfsctl ()
{
  ctltab.remove (this);
  xfree (aup.aup_machname);
  xfree (aup.aup_gids);
}

inline nfsstat3
clnt2nfs (clnt_stat err)
{
  switch (err) {
  case RPC_SUCCESS:
    return NFS3_OK;
  case RPC_CANTSEND:
  case RPC_CANTRECV:
    return NFS3ERR_JUKEBOX;
  case RPC_AUTHERROR:
    return NFS3ERR_ACCES;
  default:
    return NFS3ERR_IO;
  }
}

static void
idnames_cb (svccb *sbp, sfs_idnames *resp, clnt_stat err)
{
  sfsctl_getidnames_res res (clnt2nfs (err));
  if (!res.status)
    *res.names = *resp;
  sbp->reply (&res);
  delete resp;
}

static void
idnums_cb (svccb *sbp, sfs_idnums *resp, clnt_stat err)
{
  sfsctl_getidnums_res res (clnt2nfs (err));
  if (!res.status)
    *res.nums = *resp;
  sbp->reply (&res);
  delete resp;
}

static void
getcred_cb (svccb *sbp, sfs_cred *resp, clnt_stat err)
{
  sfsctl_getcred_res res (clnt2nfs (err));
  if (!res.status)
    *res.cred = *resp;
  sbp->reply (&res);
  delete resp;
}

static void
lookup_cb (svccb *sbp, lookup3res *resp, clnt_stat err)
{
  if (err)
    resp->set_status (clnt2nfs (err));
  sbp->reply (resp);
  delete resp;
}

static void
sfsctl_err (svccb *sbp, nfsstat3 err)
{
  switch (sbp->proc ()) {
  case SFSCTL_LOOKUP:
    sbp->replyref (lookup3res (nfsstat3 (err)));
    break;
  default:
    sbp->replyref (err);
  }
}

void
sfsctl::dispatch (svccb *sbp)
{
  if (!sbp) {
    delete this;
    return;
  }

  switch (sbp->proc ()) {
  case SFSCTL_NULL:
    sbp->reply (NULL);
    return;
  case SFSCTL_SETPID:
    setpid (*sbp->template getarg<int32_t> ());
    sbp->reply (NULL);
    return;
  }

  server *si = server::lookup (*sbp->template getarg<filename3> ());
  if (!si) {
    sfsctl_err (sbp, NFS3ERR_STALE);
    return;
  }
  if (!si->sfsc) {
    sfsctl_err (sbp, NFS3ERR_JUKEBOX);
    return;
  }
  AUTH *auth = si->getauth (aid);

  switch (sbp->proc ()) {
  case SFSCTL_GETFH:
    {
      sfsctl_getfh_res res;
      if (fip && fip->fspath == si->path)
	*res.fh = fip->fh;
      else
	res.set_status (NFS3ERR_STALE);
      fip = NULL;
      sbp->reply (&res);
      break;
    }

  case SFSCTL_GETIDNAMES:
    {
      sfsctl_getidnames_arg *argp
	= sbp->template getarg<sfsctl_getidnames_arg> ();
      sfs_idnames *resp = New sfs_idnames;
      si->sfsc->call (SFSPROC_IDNAMES, &argp->nums, resp,
		      wrap (idnames_cb, sbp, resp), auth);
      break;
    }

  case SFSCTL_GETIDNUMS:
    {
      sfsctl_getidnums_arg *argp
	= sbp->template getarg<sfsctl_getidnums_arg> ();
      sfs_idnums *resp = New sfs_idnums;
      si->sfsc->call (SFSPROC_IDNUMS, &argp->names, resp,
		      wrap (idnums_cb, sbp, resp), auth);
      break;
    }

  case SFSCTL_GETCRED:
    {
      sfs_cred *resp = New sfs_cred;
      si->sfsc->call (SFSPROC_GETCRED, NULL, resp,
		      wrap (getcred_cb, sbp, resp), auth);
      break;
    }

  case SFSCTL_LOOKUP:
    {
      sfsctl_lookup_arg *argp
	= sbp->template getarg<sfsctl_lookup_arg> ();
      lookup3res *resp = New lookup3res;
      si->sfsc->call (NFSPROC3_LOOKUP, &argp->arg, resp,
		      wrap (lookup_cb, sbp, resp), auth);
    }

  default:
    sbp->reject (PROC_UNAVAIL);
    break;
  }
}

void
sfsctl_accept (ptr<axprt_stream> x, const authunix_parms *aup)
{
  if (x && !x->ateof ())
    vNew sfsctl (x, aup);
}

static bool
mklnkfattr (fattr3exp *f, u_int32_t srvno, const nfs_fh3 *fh)
{
  bzero (f, sizeof (*f));
  f->type = NF3LNK;
  f->mode = 0444;
  f->nlink = 1;
  f->gid = sfs_gid;
  f->used = 0;
  f->fileid = 1;

  switch (srvno) {
  case srvno_fh:
    f->size = 2 * fh->data.size ();
    break;
  case srvno_fs:
    if (fh->data.size () == 4)
      if (server *s = srvnotab[getint (fh->data.base ())])
	f->size = s->path.len ();
    break;
  default:
    return false;
  }
  return true;
}

static str
mklnkdata (u_int32_t srvno, const nfs_fh3 *fh)
{
  static const char hexchars[] = "0123456789abcdef";
  switch (srvno) {
  case srvno_fh:
    {
      mstr m (2 * fh->data.size ());
      for (size_t i = 0; i < fh->data.size (); i++) {
	u_char b = fh->data[i];
	m[2*i] = hexchars[b>>4];
	m[2*i+1] = hexchars[b&0xf];
      }
      return m;
    }
  case srvno_fs:
    if (fh->data.size () == 4)
      if (server *s = srvnotab[getint (fh->data.base ())])
	return s->path;
    return NULL;
  default:
    return NULL;
  }
}

bool
sfsctl_dofakefh (u_int32_t srvno, svccb *sbp)
{
  switch (sbp->proc ()) {
  case NFSPROC3_GETATTR:
    {
      nfs_fh3 *arg = sbp->template getarg<nfs_fh3> ();
      getattr3res res (NFS3_OK);
      if (!mklnkfattr (res.attributes.addr (), srvno, arg))
	res.set_status (NFS3ERR_STALE);
      sbp->reply (&res);
      break;
    }
  case NFSPROC3_READLINK:
    {
      nfs_fh3 *arg = sbp->template getarg<nfs_fh3> ();
      readlink3res res (NFS3_OK);
      res.resok->symlink_attributes.set_present (true);
      if (!mklnkfattr (res.resok->symlink_attributes.attributes.addr (),
		       srvno, arg)
	  || !(res.resok->data = mklnkdata (srvno, arg)))
	res.set_status (NFS3ERR_STALE);
      sbp->reply (&res);
      break;
    }
  default:
    nfs3_err (sbp, NFS3ERR_ACCES);
  }
  return true;
}

bool
sfsctl_intercept (server *s, u_int32_t srvno, svccb *sbp)
{
  const authunix_parms *aup = sbp->getaup ();
  if (!aup)
    return false;
  if (!s)
    return sfsctl_dofakefh (srvno, sbp);

  switch (sbp->proc ()) {
  case NFSPROC3_SETATTR:
    {
      setattr3args *sar = sbp->template getarg<setattr3args> ();
      sattr3 &sa = sar->new_attributes;
      if (sa.mode.set || sa.size.set || sa.atime.set || sa.mtime.set
	  || !sa.uid.set || !sa.gid.set || *sa.uid.val != (u_int32_t) -2)
	return false;
      if (sfsctl *sc = ctltab (aup2aid (aup), *sa.gid.val))
	sc->fip = New refcounted<fileinfo> (s, sar->object);
      nfs3_err (sbp, NFS3ERR_PERM);
      return true;
    }
  case NFSPROC3_LOOKUP:
    {
      diropargs3 *arg = sbp->template getarg<diropargs3> ();
      if (strncmp (arg->name, SFSPREF, sizeof (SFSPREF) - 1))
	return false;
      lookup3res res (NFS3_OK);
      res.resok->obj_attributes.set_present (true);
      if (arg->name == SFSFH) {
	res.resok->object = arg->dir;
	srvno = srvno_fh;
      }
      else if (arg->name == SFSFS) {
	res.resok->object.data.setsize (4);
	putint (res.resok->object.data.base (), srvno);
	srvno = srvno_fs;
      }
      else
	return false;
      mklnkfattr (res.resok->obj_attributes.attributes.addr (),
		  srvno, &arg->dir);
      s->sendreply (sbp, &res, srvno);
      return true;
    }
  default:
    return false;
  }
}
